#!/bin/sh
#
#  link-inclibs.sh
#
#  The iODBC driver manager.
#  
#  Copyright (C) 1996-2004 by OpenLink Software <iodbc@openlinksw.com>
#  All Rights Reserved.
#
#  This software is released under the terms of either of the following
#  licenses:
#
#      - GNU Library General Public License (see LICENSE.LGPL) 
#      - The BSD License (see LICENSE.BSD).
#
#  While not mandated by the BSD license, any patches you make to the
#  iODBC source code may be contributed back into the iODBC project
#  at your discretion. Contributions will benefit the Open Source and
#  Data Access community as a whole. Submissions may be made at:
#
#      http://www.iodbc.org
#
#
#  GNU Library Generic Public License Version 2
#  ============================================
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Library General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Library General Public License for more details.
#
#  You should have received a copy of the GNU Library General Public
#  License along with this library; if not, write to the Free
#  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#
#  The BSD License
#  ===============
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. Neither the name of OpenLink Software Inc. nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
#  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#
#  Installation PATHS
#
PREFIX="/usr/local/iODBC"
ODBC_FW="/Library/Frameworks/iODBC.framework"
INST_FW="/Library/Frameworks/iODBCinst.framework"


#
#  Remove old installation
#
if [ -d "$PREFIX" ] ; then
  rm -rf "$PREFIX"
fi


#
#  Create new directory structure
#
mkdir -p "$PREFIX/bin"
mkdir -p "$PREFIX/lib"
mkdir -p "$PREFIX/include"


#
#  Create symlinks for all header files
#
ln -s "$ODBC_FW/Headers/iodbc.h"	"$PREFIX/include/iodbc.h"
ln -s "$ODBC_FW/Headers/iodbcext.h"	"$PREFIX/include/iodbcext.h"
ln -s "$ODBC_FW/Headers/iodbcunix.h"	"$PREFIX/include/iodbcunix.h"
ln -s "$ODBC_FW/Headers/isql.h"		"$PREFIX/include/isql.h"
ln -s "$ODBC_FW/Headers/isqlext.h"	"$PREFIX/include/isqlext.h"
ln -s "$ODBC_FW/Headers/isqltypes.h"	"$PREFIX/include/isqltypes.h"
ln -s "$ODBC_FW/Headers/sql.h"		"$PREFIX/include/sql.h"
ln -s "$ODBC_FW/Headers/sqlext.h"	"$PREFIX/include/sqlext.h"
ln -s "$ODBC_FW/Headers/sqltypes.h"	"$PREFIX/include/sqltypes.h"
ln -s "$ODBC_FW/Headers/sqlucode.h"	"$PREFIX/include/sqlucode.h"

ln -s "$INST_FW/Headers/iodbcinst.h"	"$PREFIX/include/iodbcinst.h"
ln -s "$INST_FW/Headers/odbcinst.h"	"$PREFIX/include/odbcinst.h"


#
#  Create symlinks for libraries
#
ln -s "$ODBC_FW/iODBC"			"$PREFIX/lib/libiodbc.dylib"
ln -s "$INST_FW/iODBCinst"		"$PREFIX/lib/libiodbcinst.dylib"


#
# Fix basic permissions
#
chmod -R og+rX "$PREFIX"


#
#  Done
#
exit 0
